import unittest

from cliotp.secret import Secret
from cliotp.totp import Totp


# https://datatracker.ietf.org/doc/html/rfc6238
# Expected values from Appendix B
class TotpTestCase(unittest.TestCase):
    def setUp(self):
        self.secret = self.pad_secret("12345678901234567890")
        self.secret_256 = self.pad_secret("12345678901234567890", 32)
        self.secret_512 = self.pad_secret("12345678901234567890", 64)

        self.times = [
            59,
            1111111109,
            1111111111,
            1234567890,
            2000000000,
            20000000000,
        ]

    def pad_secret(self, secret, length=20):
        while len(secret) < length:
            secret += secret

        return Secret.from_string(secret[:length])

    def test_sha1(self):
        expected_codes = [
            94287082,
            7081804,
            14050471,
            89005924,
            69279037,
            65353130,
        ]

        totp = Totp(self.secret)

        codes = [totp.generate_code(t) for t in self.times]

        self.assertEqual(expected_codes, codes)

    def test_sha256(self):
        expected_codes = [
            46119246,
            68084774,
            67062674,
            91819424,
            90698825,
            77737706,
        ]

        totp = Totp(self.secret_256, algorithm="sha256")

        codes = [totp.generate_code(t) for t in self.times]

        self.assertEqual(expected_codes, codes)

    def test_sha512(self):
        expected_codes = [
            90693936,
            25091201,
            99943326,
            93441116,
            38618901,
            47863826,
        ]

        totp = Totp(self.secret_512, algorithm="sha512")

        codes = [totp.generate_code(t) for t in self.times]

        self.assertEqual(expected_codes, codes)



import binascii
import os

from cryptography.hazmat.primitives.ciphers import Cipher
from cryptography.hazmat.primitives.ciphers.algorithms import AES
from cryptography.hazmat.primitives.ciphers.modes import CBC
from cryptography.hazmat.primitives.kdf.argon2 import Argon2id

from config import PASSWORD_FILE


class Crypto:
    def __init__(self, model):
        self.model = model
        self.algorithm = AES(self._master_key())

    def encrypt(self, plain_text_code):
        encryptor = self._get_cipher().encryptor()
        cipher_text = encryptor.update(plain_text_code.encode())

        return cipher_text

    def decrypt(self, cipher_text):
        encryptor = self._get_cipher().decryptor()
        plain_text = encryptor.update(cipher_text)

        return plain_text.decode()

    @staticmethod
    def save_master_password(password, salt):
        with open(PASSWORD_FILE, "a") as file:
            master_password = Crypto._generate_master_password(
                password=password, salt=salt
            )
            file.write(master_password)

    @staticmethod
    def _generate_master_password(password, salt):
        kdf = Argon2id(
            salt=salt,
            length=16,
            iterations=1,
            lanes=4,
            memory_cost=2 * 1024 * 1024,
            ad=None,
            secret=None,
        )

        digest = kdf.derive(password.encode())
        return binascii.hexlify(digest).decode()

    @staticmethod
    def random_bytes(num=16):
        return os.urandom(num)

    def _master_key(self):
        with open(PASSWORD_FILE) as file:
            contents = file.read()
            password = contents.replace("\n", "")
            return password.encode()

    def _get_cipher(self):
        return Cipher(self.algorithm, mode=CBC(self.model.initialization_vector))



# Generated by Django 5.1.6 on 2025-03-03 04:43

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("db", "0002_account_algorithm_account_digits_account_period"),
    ]

    operations = [
        migrations.AlterField(
            model_name="account",
            name="algorithm",
            field=models.CharField(
                choices=[("0", "sha1"), ("1", "sha256"), ("2", "sha512")],
                default="0",
                max_length=10,
            ),
        ),
    ]



from config import DB_PATH

# Based on https://abdus.dev/posts/django-orm-standalone/


def init_django():
    import django
    from django.conf import settings

    if settings.configured:
        return

    settings.configure(
        DEFAULT_AUTO_FIELD="django.db.models.AutoField",
        INSTALLED_APPS=[
            "db",
        ],
        DATABASES={
            "default": {
                "ENGINE": "django.db.backends.sqlite3",
                "NAME": DB_PATH,
            }
        },
    )
    django.setup()


if __name__ == "__main__":
    from django.core.management import execute_from_command_line

    init_django()
    execute_from_command_line()



import hmac
from hashlib import sha1, sha256, sha512


class Hotp:
    def __init__(self, secret, code_digits=6, algorithm="sha1"):
        algorithm = self._get_hash_algorithm(algorithm)

        self.code_digits = code_digits
        self.secret = secret.bytes
        self.hash_algorithm = algorithm

    def generate_code(self, counter):
        self._set_counter(counter)
        hs = self.hmac_sha()
        snum = self.dynamic_truncation(hs)

        return self.compute_hotp(snum)

    def hmac_sha(self):
        digester = hmac.new(self.secret, self.counter, self.hash_algorithm)
        return digester.digest()

    def dynamic_truncation(self, hmac_result):
        last_byte = hmac_result[-1]

        offset = last_byte & 0xF

        return (
            (hmac_result[offset] & 0x7F) << 24
            | (hmac_result[offset + 1] & 0xFF) << 16
            | (hmac_result[offset + 2] & 0xFF) << 8
            | (hmac_result[offset + 3] & 0xFF)
        )

    def compute_hotp(self, snum):
        return snum % 10**self.code_digits

    def _set_counter(self, counter):
        self.counter = counter.to_bytes(8, byteorder="big")

    def _get_hash_algorithm(self, algorithm):
        match algorithm.lower():
            case "sha1":
                return sha1
            case "sha256":
                return sha256
            case "sha512":
                return sha512
            case _:
                raise Exception("Invalid Algorithm")



# Generated by Django 5.1.6 on 2025-05-28 04:31

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("db", "0006_account_initialization_vector"),
    ]

    operations = [
        migrations.AlterField(
            model_name="account",
            name="encrpyted_seed",
            field=models.BinaryField(max_length=64),
        ),
        migrations.AlterField(
            model_name="account",
            name="initialization_vector",
            field=models.BinaryField(max_length=32),
        ),
    ]



# Generated by Django 5.1.6 on 2025-05-28 04:47

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("db", "0007_alter_account_encrpyted_seed_and_more"),
    ]

    operations = [
        migrations.RenameField(
            model_name="account",
            old_name="encrpyted_seed",
            new_name="encrypted_seed",
        ),
    ]



# Generated by Django 5.1.7 on 2025-05-24 00:38

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("db", "0003_alter_account_algorithm"),
    ]

    operations = [
        migrations.AddField(
            model_name="group",
            name="salt",
            field=models.BinaryField(
                default=b"2{u\x81N\x97\xe5e;\xbb\x0c\x13O\xa9\xf0\x9b", max_length=32
            ),
            preserve_default=False,
        ),
    ]



# Generated by Django 5.1.6 on 2025-03-03 04:27

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("db", "0001_initial"),
    ]

    operations = [
        migrations.AddField(
            model_name="account",
            name="algorithm",
            field=models.CharField(
                choices=[("SHA1", "sha1"), ("SHA256", "sha256"), ("SHA512", "sha512")],
                default="SHA1",
                max_length=10,
            ),
        ),
        migrations.AddField(
            model_name="account",
            name="digits",
            field=models.IntegerField(default=6),
        ),
        migrations.AddField(
            model_name="account",
            name="period",
            field=models.IntegerField(default=30),
        ),
    ]



from datetime import datetime, timezone

from .hotp import Hotp


class Totp:
    def __init__(self, secret, algorithm="sha1", time_step=30, code_digits=8):
        self.hotp = Hotp(secret, code_digits=code_digits, algorithm=algorithm)
        self.time_step = time_step

    def generate_code(self, current_timestamp=None):
        current_timestamp = current_timestamp or int(
            datetime.now(tz=timezone.utc).timestamp()
        )

        t = int(current_timestamp / self.time_step)
        return self.hotp.generate_code(t)



import unittest

from cliotp.hotp import Hotp
from cliotp.secret import Secret


# https://datatracker.ietf.org/doc/html/rfc4226
# Expected values from Appendix D
class HotpTestCase(unittest.TestCase):
    def setUp(self):
        secret = Secret.from_string("12345678901234567890")
        self.hotp = Hotp(secret)

    def test_hmac_sha(self):
        expected_hex = [
            "cc93cf18508d94934c64b65d8ba7667fb7cde4b0",
            "75a48a19d4cbe100644e8ac1397eea747a2d33ab",
            "0bacb7fa082fef30782211938bc1c5e70416ff44",
            "66c28227d03a2d5529262ff016a1e6ef76557ece",
            "a904c900a64b35909874b33e61c5938a8e15ed1c",
            "a37e783d7b7233c083d4f62926c7a25f238d0316",
            "bc9cd28561042c83f219324d3c607256c03272ae",
            "a4fb960c0bc06e1eabb804e5b397cdc4b45596fa",
            "1b3c89f65e6c9e883012052823443f048b4332db",
            "1637409809a679dc698207310c8c7fc07290d9e5",
        ]

        hex = []
        for i in range(10):
            self.hotp._set_counter(i)
            hex.append(self.hotp.hmac_sha().hex())

        self.assertEqual(expected_hex, hex)

    def test_dynamic_truncation(self):
        expected_values = [
            1284755224,
            1094287082,
            137359152,
            1726969429,
            1640338314,
            868254676,
            1918287922,
            82162583,
            673399871,
            645520489,
        ]

        values = []
        for i in range(10):
            self.hotp._set_counter(i)
            values.append(self.hotp.dynamic_truncation(self.hotp.hmac_sha()))

        self.assertEqual(expected_values, values)

    def test_generate_code(self):
        expected_codes = [
            755224,
            287082,
            359152,
            969429,
            338314,
            254676,
            287922,
            162583,
            399871,
            520489,
        ]

        codes = [self.hotp.generate_code(i) for i in range(10)]
        self.assertEqual(expected_codes, codes)



import os
import time
from datetime import datetime, timezone
from multiprocessing import Process
from pathlib import Path

import click
from django.core.management import call_command
from django.db.models import Q
from rich.console import Console
from rich.progress import BarColumn, Progress, TextColumn, TimeRemainingColumn
from rich.table import Table

from config import DB_PATH, GROUP_NAME, PASSWORD_FILE
from crypto import Crypto
from db.models import Account, Group, Tag

from .secret import Secret
from .totp import Totp


def count_down(start, code, time_step):
    with Progress(
        TextColumn(f"[bold green]{code}"),
        BarColumn(),
        TimeRemainingColumn(),
        transient=True,
    ) as progress:
        task = progress.add_task("", total=time_step, completed=start)

        while not progress.finished:
            progress.update(task, advance=1)
            time.sleep(1)


@click.group()
def cli():
    pass


@cli.command()
@click.pass_context
def init(ctx):
    click.secho(
        r"""
     ______     __         __     ______     ______   ______
    /\  ___\   /\ \       /\ \   /\  __ \   /\__  _\ /\  == \
    \ \ \____  \ \ \____  \ \ \  \ \ \/\ \  \/_/\ \/ \ \  _-/
     \ \_____\  \ \_____\  \ \_\  \ \_____\    \ \_\  \ \_\
      \/_____/   \/_____/   \/_/   \/_____/     \/_/   \/_/
    """,
        fg="magenta",
    )

    if Path(DB_PATH).is_file():
        click.secho(f"Already initialized: {DB_PATH}", fg="yellow")
        return

    call_command("migrate", "db")
    Group.objects.get_or_create(name=GROUP_NAME, salt=os.urandom(16))

    ctx.invoke(create_password)


@cli.command()
def create_password():
    if Path(PASSWORD_FILE).is_file():
        click.secho(f"Password already saved: {PASSWORD_FILE}", fg="yellow")
        return

    password = click.prompt(
        "Enter a password", type=str, confirmation_prompt=True, hide_input=True
    )
    group = Group.objects.get(name=GROUP_NAME)

    t = Process(
        target=Crypto.save_master_password,
        kwargs={"password": password, "salt": group.salt},
    )
    t.start()

    console = Console()
    with console.status("Deriving master password", spinner="bouncingBar") as status:
        status.update(status="Deriving master password")
        t.join()


@cli.command()
@click.argument("service")
@click.argument("seed")
@click.option(
    "-n",
    "--name",
    default="",
    help="Name to differentiate accounts of the same service",
)
@click.option("-t", "--tag", default=[], help="Tags to apply", multiple=True)
def add(service, seed, name, tag):
    group, _ = Group.objects.get_or_create(name=GROUP_NAME)
    click.secho(f"Adding {service}:{name}", fg="green")

    account = Account.objects.create(
        service=service,
        seed=seed,
        name=name,
        group=group,
        initialization_vector=Crypto.random_bytes(),
    )

    if tag:
        for t in tag:
            tag_object, _ = Tag.objects.get_or_create(group=group, name=t)
            tag_object.account.add(account)


@cli.command()
@click.argument("id")
def remove(id):
    group, _ = Group.objects.get_or_create(name=GROUP_NAME)
    account = group.account_set.get(id=id)

    click.secho(f"Removing {account.service}:{account.name}", fg="red")

    account.delete()


@cli.command()
@click.argument("identifier")
def code(identifier):
    group, _ = Group.objects.get_or_create(name=GROUP_NAME)
    account_set = group.account_set
    account = (
        account_set.filter(id=identifier).first()
        if identifier.isdigit()
        else group.account_set.filter(service=identifier).first()
        or group.account_set.filter(name=identifier).first()
    )

    if not account:
        return click.secho(f"No matching account found for {identifier}", fg="yellow")

    time_step = account.period
    code_length = account.digits

    totp = Totp(
        Secret.from_base32(account.seed),
        code_digits=code_length,
        algorithm=account.get_algorithm(),
        time_step=time_step,
    )

    try:
        while 1:
            start = datetime.now(tz=timezone.utc).second % time_step
            code = str(totp.generate_code()).zfill(code_length)
            count_down(start, code, time_step)
    except KeyboardInterrupt:
        pass


@cli.command()
@click.option(
    "-t",
    "--term",
    default="",
    help="Searches service, name, and tags",
)
def list(term):
    table = Table(title="Accounts", show_lines=True)
    table.add_column("ID")
    table.add_column("Service", style="cyan")
    table.add_column("Name", style="magenta")
    table.add_column("Tags")

    group, _ = Group.objects.get_or_create(name=GROUP_NAME)

    accounts = (
        group.account_set.filter(
            Q(service__contains=term)
            | Q(name__contains=term)
            | Q(tag__name__contains=term)
        )
        if term
        else group.account_set.all()
    )

    for account in accounts:
        table.add_row(str(account.id), account.service, account.name, account.tags())

    console = Console()
    console.print(table)


if __name__ == "__main__":
    cli()



# Generated by Django 5.1.6 on 2025-02-23 05:20

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Group",
            fields=[
                ("id", models.AutoField(primary_key=True, serialize=False)),
                ("created_at", models.DateTimeField(auto_now_add=True, db_index=True)),
                ("updated_at", models.DateTimeField(auto_now=True)),
                ("name", models.CharField(max_length=200)),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="Account",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=200)),
                ("service", models.CharField(max_length=200)),
                ("seed", models.CharField(max_length=256)),
                (
                    "group",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, to="db.group"
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Tag",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=200)),
                ("account", models.ManyToManyField(to="db.account")),
                (
                    "group",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, to="db.group"
                    ),
                ),
            ],
        ),
    ]



# Generated by Django 5.1.7 on 2025-05-26 15:25

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("db", "0004_group_salt"),
    ]

    operations = [
        migrations.RenameField(
            model_name="account",
            old_name="seed",
            new_name="encrpyted_seed",
        ),
    ]



from django.db import models

from crypto import Crypto
from manage import init_django

init_django()


class BaseModel(models.Model):
    id = models.AutoField(primary_key=True)
    created_at = models.DateTimeField(auto_now_add=True, db_index=True)
    updated_at = models.DateTimeField(auto_now=True)

    class Meta:
        abstract = True


class Group(BaseModel):
    name = models.CharField(max_length=200)
    salt = models.BinaryField(max_length=32)


class Account(models.Model):
    SHA1 = "0"
    SHA256 = "1"
    SHA512 = "2"

    ALGORITHM_CHOICES = {
        SHA1: "sha1",
        SHA256: "sha256",
        SHA512: "sha512",
    }

    group = models.ForeignKey(Group, on_delete=models.CASCADE)
    name = models.CharField(max_length=200)
    service = models.CharField(max_length=200)
    encrypted_seed = models.BinaryField(max_length=64)
    initialization_vector = models.BinaryField(max_length=32)
    period = models.IntegerField(default=30)
    digits = models.IntegerField(default=6)
    algorithm = models.CharField(
        max_length=10,
        choices=ALGORITHM_CHOICES,
        default=SHA1,
    )

    @property
    def seed(self):
        crypto = Crypto(self)
        return crypto.decrypt(self.encrypted_seed)

    @seed.setter
    def seed(self, plaintext):
        crypto = Crypto(self)
        self.encrypted_seed = crypto.encrypt(plaintext)
        return self.encrypted_seed

    def get_algorithm(self):
        return self.get_algorithm_display()

    def tags(self):
        account_tags = self.tag_set.all()
        return ", ".join([tag.name for tag in account_tags])


class Tag(models.Model):
    group = models.ForeignKey(Group, on_delete=models.CASCADE)
    account = models.ManyToManyField(Account)
    name = models.CharField(max_length=200)



# Generated by Django 5.1.6 on 2025-05-27 03:43

from django.db import migrations, models

import crypto


class Migration(migrations.Migration):
    dependencies = [
        ("db", "0005_rename_seed_account_encrpyted_seed"),
    ]

    operations = [
        migrations.AddField(
            model_name="account",
            name="initialization_vector",
            field=models.BinaryField(default=crypto.Crypto.random_bytes, max_length=32),
        ),
    ]



import os
from pathlib import Path

from dotenv import load_dotenv

load_dotenv()

GROUP_NAME = "default"

STORAGE_PATH = Path(os.getenv("CLIOTP_STORAGE_PATH", Path.home()))
STORAGE_DIR = STORAGE_PATH.joinpath(".cliotp")

PASSWORD_FILE = STORAGE_DIR.joinpath("master-password.txt")

DB_PATH = str(STORAGE_DIR / "cliopt.sqlite3")

STORAGE_DIR.mkdir(exist_ok=True)
